#  tests for python-kaleido-0.2.1-py310haa95532_0 (this is a generated file);
print('===== testing package: python-kaleido-0.2.1-py310haa95532_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
from kaleido.scopes.plotly import PlotlyScope
import subprocess


subprocess.run(["pip", "check"], check=True)

scope = PlotlyScope()
assert scope.transform({"data": []}).startswith(b"\x89PNG")
#  --- run_test.py (end) ---

print('===== python-kaleido-0.2.1-py310haa95532_0 OK =====');
